package flink.zanxiangnet.ad.monitoring.process;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfHourODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.PlanStatOfHourDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.util.Collector;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Slf4j
public class PlanHourDWDProcess extends ProcessWindowFunction<AdDataOfHourODS, PlanStatOfHourDWD, Long, GlobalWindow> {
    
    private Odps odps;
    // 上次查询的天数据
    private ValueState<String> lastQueryDayState;
    // 聚合的天的数据
    private MapState<String, PlanStatOfHourDWD> historyReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();
        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
        odps = new Odps(account);
        odps.getRestClient().setRetryLogger(new MaxComputeLog());
        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", Types.STRING, Types.POJO(PlanStatOfHourDWD.class)));
    }

    @Override
    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfHourODS, PlanStatOfHourDWD, Long, GlobalWindow>.Context context,
                        Iterable<AdDataOfHourODS> iterable, Collector<PlanStatOfHourDWD> collector) throws Exception {
        AdDataOfHourODS element = iterable.iterator().next();
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        LocalDateTime statTime = LocalDateTime.of(statDay, LocalTime.of(element.getHour(), 0, 0));
        long now = System.currentTimeMillis();
        LocalDate today = LocalDate.now();

        String lastQueryDay = lastQueryDayState.value();
        // 从 maxCompute拉取指定 广告的历史数据
        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(today))) {
            LocalDate endDay = today, beginDay = statDay.minusDays(60);
            String sql = "SELECT * FROM plan_stat_of_hour_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endDay) + "\" AND campaign_id = " + element.getCampaignId() + ";";
            Instance instance = SQLTask.run(odps, sql);
            System.out.println("178===>sql: " + sql + ", odps日志: " + odps.logview().generateLogView(instance, 7 * 24));
            instance.waitForSuccess();
            List<Record> records = SQLTask.getResult(instance);
            Map<String, PlanStatOfHourDWD> historyHourMap = records.stream()
                    .map(PlanStatOfHourDWD::byMaxCompute)
                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                    .collect(Collectors.toMap(data -> data.getStatDay() + data.getHour(), data -> data, (val1, val2) -> val2));
            historyReduceState.clear();
            historyReduceState.putAll(historyHourMap);
            lastQueryDayState.update(DateUtil.formatLocalDate(today));
        }
        PlanStatOfHourDWD lastReduceData = null;
        for (int i = 1; i < 60 * 24; i++) {
            LocalDateTime time = statTime.minusHours(i);
            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(time.toLocalDate()) + time.getHour());
            if (lastReduceData != null) {
                break;
            }
        }

        PlanStatOfHourDWD newStatData = PlanStatOfHourDWD.reduce(lastReduceData, element, now);
        historyReduceState.put(newStatData.getStatDay(), newStatData);
        collector.collect(newStatData);
    }
}
